from openapi_core.wrappers.base import BaseOpenAPIResponse


class ClientOpenAPIResponse(BaseOpenAPIResponse):

    def __init__(self, response):
        self.response = response

    @property
    def body(self):
        return self.response.content

    @property
    def data(self):
        return self.body

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def mimetype(self):
        return self.response.headers.get('content-type')

    @property
    def payload(self):
        return self.response.json()



from functools import wraps
from importlib import import_module
from inspect import iscoroutinefunction
from pathlib import Path
from types import ModuleType
from typing import Callable, Union
from urllib.parse import urlsplit

from openapi_core import create_spec
from openapi_core.schema.specs.models import Spec
from openapi_core.shortcuts import RequestValidator, ResponseValidator
from starlette.applications import Starlette
from starlette.responses import JSONResponse, Response
from stringcase import snakecase

from pyotr.utils import get_spec_from_file
from pyotr.validation.requests import StarletteOpenAPIRequest
from pyotr.validation.responses import StarletteOpenAPIResponse


class Application(Starlette):

    def __init__(
        self, spec: Union[Spec, dict], base: Union[str, ModuleType], *,
        validate_responses: bool = True, ignore_server_paths: bool = False, **kwargs
    ):
        super().__init__(**kwargs)
        if not isinstance(spec, Spec):
            spec = create_spec(spec)
        self.spec = spec
        self.validate_responses = validate_responses
        server_paths = {''}
        if not ignore_server_paths:
            server_paths = {
                urlsplit(server.url).path for server in self.spec.servers
            }

        if isinstance(base, str):
            base = _load_module(base)

        for path, path_spec in spec.paths.items():
            for method, operation in path_spec.operations.items():
                endpoint = self._get_endpoint(operation.operation_id, base)
                for server_path in server_paths:
                    self.add_route(server_path + path, endpoint, [method])

    def _get_endpoint(self, name: str, module: ModuleType, enforce_case: bool = True) -> Callable:
        if '.' in name:
            base, name = name.rsplit('.', 1)
            module = _load_module(f'{module.__name__}.{base}')
        if enforce_case:
            name = snakecase(name)
        try:
            endpoint = getattr(module, name)
        except AttributeError as e:
            raise RuntimeError(f'The function `{module}.{name}` does not exist!') from e

        @wraps(endpoint)
        async def wrapper(request, **kwargs) -> Response:
            validation_request = await StarletteOpenAPIRequest.prepare(request)
            request_validation = RequestValidator(self.spec).validate(validation_request)
            request_validation.raise_for_errors()

            if iscoroutinefunction(endpoint):
                response = await endpoint(request, **kwargs)
            else:
                response = endpoint(request, **kwargs)
            if isinstance(response, dict):
                response = JSONResponse(response)

            # TODO: pass a list of endpoint names to specify which responses to skip
            if self.validate_responses:
                ResponseValidator(self.spec).validate(
                    validation_request, StarletteOpenAPIResponse(response)
                ).raise_for_errors()
            return response

        return wrapper

    @classmethod
    def from_file(cls, path: Union[Path, str], *args, **kwargs) -> 'Application':
        spec = get_spec_from_file(path)
        return cls(spec, *args, **kwargs)


def _load_module(name: str) -> ModuleType:
    try:
        module = import_module(name)
    except ModuleNotFoundError as e:
        raise RuntimeError(f'The module `{name}` does not exist!') from e
    else:
        return module



from string import Formatter
from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit

from openapi_core.schema.operations.models import Operation
from openapi_core.wrappers.base import BaseOpenAPIRequest


class ClientOpenAPIRequest(BaseOpenAPIRequest):

    def __init__(self, host_url: str, op_spec: Operation):
        self.spec = op_spec
        self.method = op_spec.http_method.lower()

        self.host_url = host_url
        self._url_parts = urlsplit(self.host_url)

        self.path_pattern = self._url_parts.path + op_spec.path_name

        self.parameters = {
            'path': {},
            'query': parse_qs(self._url_parts.query),
            'header': {},
            'cookie': {},
        }
        self.body = None

        self.mimetype = NotImplemented

        formatter = Formatter()
        self.url_vars = [
            var for _, var, _, _
            in formatter.parse(op_spec.path_name)
            if var is not None
        ]

    @property
    def url(self):
        url_parts = self._url_parts._asdict()
        url_parts['path'] = self.path
        url_parts['query'] = urlencode(self.parameters['query'])
        return urlunsplit(url_parts.values())

    @property
    def path(self):
        return self.path_pattern.format(**self.parameters['path'])

    @property
    def headers(self):
        return self.parameters['header']

    def prepare(self, *args, **kwargs):
        len_vars = len(self.url_vars)
        if len(args) != len_vars:
            error_message = f"Incorrect arguments: {self.spec.operation_id} accepts"
            if len_vars:
                error_message += (
                    f" {len_vars} positional argument{'s' if len_vars > 1 else ''}:"
                    f" {', '.join(self.url_vars)}"
                )
            else:
                error_message += f" no positional arguments"
            raise RuntimeError(error_message)

        self.body = kwargs.pop('data_', None)

        headers = kwargs.pop('headers_', None)
        self.parameters.update({
            'path': dict(zip(self.url_vars, args)),
            'query': kwargs,
            'header': headers
        })

        return self



import json
from enum import Enum
from itertools import chain
from pathlib import Path
from typing import Any, Union

import yaml
from openapi_core.schema.specs.models import Spec
from typing_extensions import Protocol


class SpecFileTypes(Enum):
    JSON = ('json',)
    YAML = ('yaml', 'yml')


def get_spec_from_file(path: Union[Path, str]) -> Spec:
    path = Path(path)
    suffix = path.suffix[1:].lower()
    with open(path) as spec_file:
        if suffix in SpecFileTypes.JSON.value:
            spec = json.load(spec_file)
        elif suffix in SpecFileTypes.YAML.value:
            spec = yaml.safe_load(spec_file)
        else:
            raise RuntimeError(
                f"Unknown specification file type."
                f" Accepted types: {', '.join(chain(*(i.value for i in SpecFileTypes)))}"
            )
    return spec


class Requestable(Protocol):  # pragma: no cover
    """ Any object (usually a module, class or instance) that implements the
        `request` method compatible with the `requests` library.
    """
    def request(self, method: str, url: str, **kwargs) -> Any:
        ...



from pathlib import Path
from typing import Optional, Type, Union

import httpx
from openapi_core import create_spec
from openapi_core.schema.servers.models import Server
from openapi_core.schema.specs.models import Spec
from openapi_core.shortcuts import ResponseValidator
from stringcase import snakecase

from pyotr.utils import Requestable
from pyotr.utils import get_spec_from_file
from pyotr.validation.requests import ClientOpenAPIRequest
from pyotr.validation.responses import ClientOpenAPIResponse


class Client:

    def __init__(self, spec: Union[Spec, dict], *,
                 server_url: Optional[str] = None,
                 client: Requestable = httpx,
                 request_class: Type[ClientOpenAPIRequest] = ClientOpenAPIRequest,
                 response_class: Type[ClientOpenAPIResponse] = ClientOpenAPIResponse
                 ):
        if not isinstance(spec, Spec):
            spec = create_spec(spec)
        self.spec = spec
        self.client = client
        self.request_class = request_class
        self.response_class = response_class

        if server_url is None:
            server_url = self.spec.servers[0].url
        else:
            server_url = server_url.rstrip('/')
            for server in self.spec.servers:
                if server_url == server.url:
                    break
            else:
                self.spec.servers.append(Server(server_url))
        self.server_url = server_url
        self.validator = ResponseValidator(self.spec)

        for path_spec in spec.paths.values():
            for op_spec in path_spec.operations.values():
                setattr(self, snakecase(op_spec.operation_id), self._get_operation(op_spec).__get__(self))

    @staticmethod
    def _get_operation(op_spec):
        # TODO: extract args and kwargs from operation parameters
        def operation(self, *args,
                      body_: Optional[Union[dict, list]] = None,
                      headers_: Optional[dict] = None,
                      **kwargs
                      ):
            request = self.request_class(self.server_url, op_spec)
            request.prepare(*args, data_=body_, headers_=headers_, **kwargs)
            api_response = self.client.request(
                method=request.method, url=request.url, data=request.body, headers=request.headers
            )
            api_response.raise_for_status()
            response = self.response_class(api_response)
            self.validator.validate(request, response).raise_for_errors()
            return response
        return operation

    @classmethod
    def from_file(cls, path: Union[Path, str], **kwargs):
        spec = get_spec_from_file(path)
        return cls(spec, **kwargs)
